@@ -16,6 +16,7 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#include "gnome-network-displays-config.h"
 #include "cc-ctrl.h"
 #include "cc-comm.h"
 
@@ -48,15 +49,20 @@ cc_ctrl_send_auth (CcCtrl *ctrl)
 static gboolean
 cc_ctrl_send_connect (CcCtrl *ctrl, gchar *destination_id)
 {
+#ifdef __aarch64__
+  gchar *platform = "Wayland; Linux aarch64";
+#else
+  gchar *platform = "Wayland; Linux x86_64";
+#endif
   gchar *json = cc_json_helper_build_string (
     "type", CC_JSON_TYPE_STRING, "CONNECT",
-    "userAgent", CC_JSON_TYPE_STRING, "GND/0.91.0  (X11; Linux x86_64)",
+    "userAgent", CC_JSON_TYPE_STRING, g_strdup_printf ("GND/%s (%s)", PACKAGE_VERSION, platform),
     "connType", CC_JSON_TYPE_INT, 0,
     "origin", CC_JSON_TYPE_OBJECT, cc_json_helper_build_node (NULL),
     "senderInfo", CC_JSON_TYPE_OBJECT, cc_json_helper_build_node (
       "sdkType", CC_JSON_TYPE_INT, 2,
-      "version", CC_JSON_TYPE_STRING, "X11; Linux x86_64",
-      "browserVersion", CC_JSON_TYPE_STRING, "X11; Linux x86_64",
+      "version", CC_JSON_TYPE_STRING, platform,
+      "browserVersion", CC_JSON_TYPE_STRING, platform,
       "platform", CC_JSON_TYPE_INT, 6,
       "connectionType", CC_JSON_TYPE_INT, 1,
       NULL),
@@ -208,9 +214,15 @@ cc_ctrl_send_close_app (CcCtrl *ctrl, gchar *sessionId)
 static gboolean
 cc_ctrl_send_load (CcCtrl *ctrl, gchar *sessionId)
 {
+  guint port;
+  CcMediaFactory *factory;
+
+  g_object_get (ctrl->http_server, "port", &port, NULL);
+  factory = (CcMediaFactory *) ctrl->http_server;
+
   GArray *tracks = g_array_new (FALSE, FALSE, sizeof (JsonNode *));
   JsonNode *track_node = cc_json_helper_build_node (
-    "trackContentType", CC_JSON_TYPE_STRING, "video/mp2t",
+    "trackContentType", CC_JSON_TYPE_STRING, content_types[cc_media_factory_profiles[factory->factory_profile].muxer],
     "trackId", CC_JSON_TYPE_INT, 1,
     "type", CC_JSON_TYPE_STRING, "VIDEO",
     NULL);
@@ -222,9 +234,9 @@ cc_ctrl_send_load (CcCtrl *ctrl, gchar *sessionId)
     "media", CC_JSON_TYPE_OBJECT, cc_json_helper_build_node (
       "contentUrl", CC_JSON_TYPE_STRING, g_strdup_printf ("http://%s:%d/",
                                                           ctrl->comm.local_address,
-                                                          cc_http_server_get_port (ctrl->http_server)),
+                                                          port),
       "streamType", CC_JSON_TYPE_STRING, "LIVE",
-      "contentType", CC_JSON_TYPE_STRING, "video/x-matroska",
+      "contentType", CC_JSON_TYPE_STRING, content_types[cc_media_factory_profiles[factory->factory_profile].muxer],
       NULL),
     "requestId", CC_JSON_TYPE_INT, ctrl->request_id++,
     NULL);
@@ -328,24 +340,26 @@ cc_ctrl_handle_receiver_status (CcCtrl *ctrl, JsonParser *parser)
 {
   g_autoptr(JsonNode) app_status = NULL;
   g_autoptr(JsonPath) path = json_path_new ();
-  json_path_compile (path, "$.status.applications[0]", NULL);
+
+  gchar *parsed_payload;
+
+  json_path_compile (path, "$.status.applications", NULL);
   app_status = json_path_match (path, json_parser_get_root (parser));
 
   g_autoptr(JsonGenerator) generator = json_generator_new ();
   json_generator_set_root (generator, app_status);
-  gsize size;
 
-  json_generator_to_data (generator, &size);
+  parsed_payload = json_generator_to_data (generator, NULL);
 
-  if (size == 2) /* empty array [] */
+  /* applications key not found = []
+   * applications key found but empty = [[]]
+   */
+  if (g_str_equal (parsed_payload, "[]") || g_str_equal (parsed_payload, "[[]]"))
     {
       g_debug ("CcCtrl: No apps open");
-      if (ctrl->state == CC_CTRL_STATE_LAUNCH_SENT)
+      if (ctrl->state >= CC_CTRL_STATE_LAUNCH_SENT)
         return;
 
-      if (ctrl->state >= CC_CTRL_STATE_APP_OPEN) /* app closed unexpectedly */
-        g_debug ("CcCtrl: App closed unexpectedly");
-
       if (!cc_ctrl_send_launch_app (ctrl, CC_APP_ID))
         return;
 
@@ -356,7 +370,7 @@ cc_ctrl_handle_receiver_status (CcCtrl *ctrl, JsonParser *parser)
   /* one or more apps is/are open */
   g_autoptr(JsonReader) reader = json_reader_new (app_status);
 
-  if (json_reader_read_element (reader, 0))
+  if (json_reader_read_element (reader, 0) && json_reader_read_element (reader, 0))
     {
       if (json_reader_read_member (reader, "appId"))
         {
@@ -379,22 +393,15 @@ cc_ctrl_handle_receiver_status (CcCtrl *ctrl, JsonParser *parser)
 
                   if (!cc_ctrl_app_init (ctrl, &error))
                     {
-                      GError *error_;
-                      if (error)
+                      if (!error)
                         {
-                          error_ = g_error_new (CC_ERROR,
-                                                CC_ERROR_HTTP_SERVER_LISTEN_FAILED,
-                                                "Failed to start the HTTP server");
+                          error = g_error_new (CC_ERROR,
+                                               CC_ERROR_APP_COMMUNICATION_FAILED,
+                                               "Could not initialize \"%s\" app",
+                                               CC_APP_ID);
                         }
-                      else
-                        {
-                          error_ = g_error_new (CC_ERROR,
-                                                CC_ERROR_APP_COMMUNICATION_FAILED,
-                                                "Could not initialize \"%s\" app",
-                                                CC_APP_ID);
-                        }
-                      g_warning ("CcCtrl: %s", error_->message);
-                      cc_ctrl_close_connection (ctrl, error_);
+                      g_warning ("CcCtrl: %s", error->message);
+                      cc_ctrl_close_connection (ctrl, error);
                       return;
                     }
 
@@ -446,6 +453,13 @@ cc_ctrl_handle_received_msg (gpointer                    userdata,
   g_autoptr(JsonReader) reader = NULL;
   CcReceivedMessageTypeEnum type;
 
+  if (message->payload_utf8 == NULL)
+    {
+      g_warning ("CcCtrl: Received message with no payload");
+      cc_json_helper_dump_message (message, TRUE);
+      return;
+    }
+
   parser = json_parser_new ();
   if (!json_parser_load_from_data (parser, message->payload_utf8, -1, &error))
     {
@@ -458,7 +472,7 @@ cc_ctrl_handle_received_msg (gpointer                    userdata,
   reader = json_reader_new (json_parser_get_root (parser));
   type = cc_json_helper_get_message_type (message, reader);
 
-  if (!(type == CC_REC_MSG_TYPE_PING || type == CC_REC_MSG_TYPE_PONG || type == -1))
+  if (!(type == CC_REC_MSG_TYPE_PING || type == CC_REC_MSG_TYPE_PONG))
     {
       g_debug ("CcComm: Received message:");
       cc_json_helper_dump_message (message, FALSE);
@@ -624,7 +638,3 @@ cc_ctrl_finish (CcCtrl *ctrl)
 
   ctrl->state = CC_CTRL_STATE_DISCONNECTED;
 }
-
-/* TODO: don't forget to update the state at proper places, at least check once
- * and implement all the checks wherever necessary
- */
