@@ -17,11 +17,12 @@
  */
 
 #include "gnome-network-displays-config.h"
-#include "nd-wfd-p2p-sink.h"
-#include "wfd/wfd-server.h"
-#include "wfd/wfd-client.h"
-#include "wfd/wfd-media-factory.h"
+#include "nd-enum-types.h"
 #include "nd-firewalld.h"
+#include "nd-uri-helpers.h"
+#include "nd-wfd-p2p-sink.h"
+#include "wfd/wfd-client.h"
+#include "wfd/wfd-server.h"
 
 struct _NdWFDP2PSink
 {
@@ -36,8 +37,10 @@ struct _NdWFDP2PSink
   NMWifiP2PPeer      *nm_peer;
   NMActiveConnection *nm_ac;
 
-  GStrv               missing_video_codec;
-  GStrv               missing_audio_codec;
+  gchar              *uuid;
+
+  GtkStringList      *missing_video_codec;
+  GtkStringList      *missing_audio_codec;
   char               *missing_firewall_zone;
 
   WfdServer          *server;
@@ -49,20 +52,25 @@ enum {
   PROP_DEVICE,
   PROP_PEER,
 
+  PROP_UUID,
   PROP_DISPLAY_NAME,
   PROP_MATCHES,
   PROP_PRIORITY,
   PROP_STATE,
+  PROP_PROTOCOL,
   PROP_MISSING_VIDEO_CODEC,
   PROP_MISSING_AUDIO_CODEC,
   PROP_MISSING_FIREWALL_ZONE,
 
-  PROP_LAST = PROP_DISPLAY_NAME,
+  PROP_LAST = PROP_UUID,
 };
 
+const static NdSinkProtocol protocol = ND_SINK_PROTOCOL_WFD_P2P;
+
 static void nd_wfd_p2p_sink_sink_iface_init (NdSinkIface *iface);
 static NdSink * nd_wfd_p2p_sink_sink_start_stream (NdSink *sink);
 static void nd_wfd_p2p_sink_sink_stop_stream (NdSink *sink);
+static gchar * nd_wfd_p2p_sink_sink_to_uri (NdSink *sink);
 
 static void nd_wfd_p2p_sink_sink_stop_stream_int (NdWFDP2PSink *self);
 
@@ -79,7 +87,7 @@ static void
 peer_notify_cb (NdWFDP2PSink *self, GParamSpec *pspec, NMWifiP2PPeer *peer)
 {
   /* TODO: Assumes the display name may have changed.
-   *       This is obviously overly agressive, on the other hand
+   *       This is obviously overly aggressive, on the other hand
    *       not really an issue. */
   g_object_notify (G_OBJECT (self), "display-name");
 }
@@ -124,6 +132,10 @@ nd_wfd_p2p_sink_get_property (GObject    *object,
       g_value_set_object (value, sink->nm_peer);
       break;
 
+    case PROP_UUID:
+      g_value_set_string (value, sink->uuid);
+      break;
+
     case PROP_DISPLAY_NAME:
       g_object_get_property (G_OBJECT (sink->nm_peer), "name", value);
       break;
@@ -131,9 +143,19 @@ nd_wfd_p2p_sink_get_property (GObject    *object,
     case PROP_MATCHES:
       {
         g_autoptr(GPtrArray) res = NULL;
+        const char *hw_addr;
         res = g_ptr_array_new_with_free_func (g_free);
 
-        g_ptr_array_add (res, g_strdup (nm_wifi_p2p_peer_get_hw_address (sink->nm_peer)));
+        /* Should not usually happen, but it can if something is holding on
+         * to the sink. So guard against NULL being returned if the peer
+         * object is not valid anymore. */
+        hw_addr = nm_wifi_p2p_peer_get_hw_address (sink->nm_peer);
+        if (hw_addr)
+          {
+            gchar *addr = g_utf8_strup (hw_addr, -1);
+            g_debug ("NdWFDP2PSink: Adding P2P MAC %s to match list", addr);
+            g_ptr_array_add (res, addr);
+          }
 
         g_value_take_boxed (value, g_steal_pointer (&res));
         break;
@@ -147,12 +169,16 @@ nd_wfd_p2p_sink_get_property (GObject    *object,
       g_value_set_enum (value, sink->state);
       break;
 
+    case PROP_PROTOCOL:
+      g_value_set_enum (value, protocol);
+      break;
+
     case PROP_MISSING_VIDEO_CODEC:
-      g_value_set_boxed (value, sink->missing_video_codec);
+      g_value_set_object (value, sink->missing_video_codec);
       break;
 
     case PROP_MISSING_AUDIO_CODEC:
-      g_value_set_boxed (value, sink->missing_audio_codec);
+      g_value_set_object (value, sink->missing_audio_codec);
       break;
 
     case PROP_MISSING_FIREWALL_ZONE:
@@ -213,17 +239,17 @@ nd_wfd_p2p_sink_finalize (GObject *object)
 {
   NdWFDP2PSink *sink = ND_WFD_P2P_SINK (object);
 
+  nd_wfd_p2p_sink_sink_stop_stream_int (sink);
+
   g_cancellable_cancel (sink->cancellable);
   g_clear_object (&sink->cancellable);
 
-  nd_wfd_p2p_sink_sink_stop_stream_int (sink);
-
   g_clear_object (&sink->nm_client);
   g_clear_object (&sink->nm_device);
   g_clear_object (&sink->nm_peer);
 
-  g_clear_pointer (&sink->missing_video_codec, g_strfreev);
-  g_clear_pointer (&sink->missing_audio_codec, g_strfreev);
+  g_clear_object (&sink->missing_video_codec);
+  g_clear_object (&sink->missing_audio_codec);
   g_clear_pointer (&sink->missing_firewall_zone, g_free);
 
   G_OBJECT_CLASS (nd_wfd_p2p_sink_parent_class)->finalize (object);
@@ -258,10 +284,12 @@ nd_wfd_p2p_sink_class_init (NdWFDP2PSinkClass *klass)
 
   g_object_class_install_properties (object_class, PROP_LAST, props);
 
+  g_object_class_override_property (object_class, PROP_UUID, "uuid");
   g_object_class_override_property (object_class, PROP_DISPLAY_NAME, "display-name");
   g_object_class_override_property (object_class, PROP_MATCHES, "matches");
   g_object_class_override_property (object_class, PROP_PRIORITY, "priority");
   g_object_class_override_property (object_class, PROP_STATE, "state");
+  g_object_class_override_property (object_class, PROP_PROTOCOL, "protocol");
   g_object_class_override_property (object_class, PROP_MISSING_VIDEO_CODEC, "missing-video-codec");
   g_object_class_override_property (object_class, PROP_MISSING_AUDIO_CODEC, "missing-audio-codec");
   g_object_class_override_property (object_class, PROP_MISSING_FIREWALL_ZONE, "missing-firewall-zone");
@@ -270,10 +298,31 @@ nd_wfd_p2p_sink_class_init (NdWFDP2PSinkClass *klass)
 static void
 nd_wfd_p2p_sink_init (NdWFDP2PSink *sink)
 {
+  sink->uuid = g_uuid_string_random ();
   sink->state = ND_SINK_STATE_DISCONNECTED;
   sink->cancellable = g_cancellable_new ();
 }
 
+static gchar *
+nd_wfd_p2p_sink_sink_to_uri (NdSink *sink)
+{
+  NdWFDP2PSink *self = ND_WFD_P2P_SINK (sink);
+  GHashTable *params = g_hash_table_new (g_str_hash, g_str_equal);
+
+  /* protocol */
+  g_hash_table_insert (params, "protocol", (gpointer *) g_strdup_printf ("%d", protocol));
+
+  /* device */
+  const gchar *device_path = nm_object_get_path ((NMObject *) self->nm_device);
+  g_hash_table_insert (params, "device", (gpointer *) g_strdup (device_path));
+
+  /* peer */
+  const gchar *peer_path = nm_object_get_path ((NMObject *) self->nm_peer);
+  g_hash_table_insert (params, "peer", (gpointer *) g_strdup (peer_path));
+
+  return nd_uri_helpers_generate_uri (params);
+}
+
 /******************************************************************
 * NdSink interface implementation
 ******************************************************************/
@@ -283,6 +332,7 @@ nd_wfd_p2p_sink_sink_iface_init (NdSinkIface *iface)
 {
   iface->start_stream = nd_wfd_p2p_sink_sink_start_stream;
   iface->stop_stream = nd_wfd_p2p_sink_sink_stop_stream;
+  iface->to_uri = nd_wfd_p2p_sink_sink_to_uri;
 }
 
 static void
@@ -350,8 +400,10 @@ p2p_connected (GObject      *source_object,
                gpointer      user_data)
 {
   g_autoptr(GError) error = NULL;
-  NdWFDP2PSink *sink = NULL;
+  NdWFDP2PSink *sink = ND_WFD_P2P_SINK (user_data);
   NMActiveConnection *ac = NULL;
+  GStrv missing_video = NULL, missing_audio = NULL;
+  gboolean have_wfd_codecs;
 
   g_debug ("NdWfdP2PSink: Got P2P connection");
 
@@ -363,18 +415,39 @@ p2p_connected (GObject      *source_object,
         return;
 
       g_warning ("Error activating connection: %s", error->message);
-      sink = ND_WFD_P2P_SINK (user_data);
       nd_wfd_p2p_sink_sink_stop_stream_int (sink);
       sink->state = ND_SINK_STATE_ERROR;
       g_object_notify (G_OBJECT (sink), "state");
       return;
     }
 
-  sink = ND_WFD_P2P_SINK (user_data);
   sink->nm_ac = ac;
 
   g_assert (sink->server == NULL);
   sink->server = wfd_server_new ();
+
+  have_wfd_codecs = wfd_server_lookup_encoders (sink->server,
+                                                &missing_video,
+                                                &missing_audio);
+
+  g_clear_object (&sink->missing_video_codec);
+  g_clear_object (&sink->missing_audio_codec);
+
+  sink->missing_video_codec = gtk_string_list_new ((const char *const *) missing_video);
+  sink->missing_audio_codec = gtk_string_list_new ((const char *const *) missing_audio);
+
+  g_object_notify (G_OBJECT (sink), "missing-video-codec");
+  g_object_notify (G_OBJECT (sink), "missing-audio-codec");
+
+  if (!have_wfd_codecs)
+    {
+      g_warning ("NdWfdP2PSink: Essential codecs are missing!");
+      sink->state = ND_SINK_STATE_ERROR;
+      g_object_notify (G_OBJECT (sink), "state");
+
+      return;
+    }
+
   /*
    * XXX: Not yet implemented, but we should only bind on the P2P device
    * wfd_server_set_interface (GST_RTSP_SERVER (sink->server), nm_device_get_ip_iface (sink->nm_device));
@@ -466,6 +539,7 @@ firewall_ready (GObject      *source_object,
   connection = nm_simple_connection_new ();
 
   general_setting = nm_setting_connection_new ();
+  nm_setting_connection_add_permission ((NMSettingConnection *) general_setting, "user", g_get_user_name (), NULL);
   nm_connection_add_setting (connection, general_setting);
   g_object_set (general_setting, NM_SETTING_CONNECTION_ZONE, ND_WFD_ZONE, NULL);
 
@@ -482,7 +556,7 @@ firewall_ready (GObject      *source_object,
                 NULL);
 
   /* We do not need IPv6 */
-  ipv6_setting = nm_setting_ip4_config_new ();
+  ipv6_setting = nm_setting_ip6_config_new ();
   nm_connection_add_setting (connection, ipv6_setting);
   g_object_set (ipv6_setting,
                 NM_SETTING_IP_CONFIG_METHOD, NM_SETTING_IP6_CONFIG_METHOD_AUTO,
@@ -503,36 +577,11 @@ firewall_ready (GObject      *source_object,
 static NdSink *
 nd_wfd_p2p_sink_sink_start_stream (NdSink *sink)
 {
+  g_autoptr(NdFirewalld) firewalld = NULL;
   NdWFDP2PSink *self = ND_WFD_P2P_SINK (sink);
 
-  g_autoptr(NdFirewalld) firewalld = NULL;
-  gboolean have_basic_codecs;
-  GStrv missing_video, missing_audio;
-
   g_return_val_if_fail (self->state == ND_SINK_STATE_DISCONNECTED, NULL);
 
-  g_assert (self->server == NULL);
-
-  have_basic_codecs = wfd_get_missing_codecs (&missing_video, &missing_audio);
-
-  g_clear_pointer (&self->missing_video_codec, g_strfreev);
-  g_clear_pointer (&self->missing_audio_codec, g_strfreev);
-
-  self->missing_video_codec = g_strdupv (missing_video);
-  self->missing_audio_codec = g_strdupv (missing_audio);
-
-  g_object_notify (G_OBJECT (self), "missing-video-codec");
-  g_object_notify (G_OBJECT (self), "missing-audio-codec");
-
-  if (!have_basic_codecs)
-    {
-      g_warning ("Essential codecs are missing!");
-      self->state = ND_SINK_STATE_ERROR;
-      g_object_notify (G_OBJECT (self), "state");
-
-      return g_object_ref (sink);
-    }
-
   self->state = ND_SINK_STATE_ENSURE_FIREWALL;
   g_object_notify (G_OBJECT (self), "state");
 
@@ -574,7 +623,7 @@ nd_wfd_p2p_sink_sink_stop_stream_int (NdWFDP2PSink *self)
    * nm_ac will be unset if something else destroyed the connection already */
   if (self->nm_ac)
     {
-      nm_device_disconnect (self->nm_device, NULL, NULL);
+      nm_device_disconnect_async (self->nm_device, NULL, NULL, NULL);
       g_clear_object (&self->nm_ac);
     }
 }
@@ -636,6 +685,20 @@ nd_wfd_p2p_sink_get_peer (NdWFDP2PSink * sink)
   return sink->nm_peer;
 }
 
+/**
+ * nd_wfd_p2p_sink_get_state
+ * @sink: a #NdWFDP2PSink
+ *
+ * Retrieve the #NdSinkState.
+ *
+ * Returns: (transfer none): The #NdSinkState
+ */
+NdSinkState
+nd_wfd_p2p_sink_get_state (NdWFDP2PSink *sink)
+{
+  return sink->state;
+}
+
 NdWFDP2PSink *
 nd_wfd_p2p_sink_new (NMClient *client, NMDevice *device, NMWifiP2PPeer * peer)
 {
@@ -645,3 +708,68 @@ nd_wfd_p2p_sink_new (NMClient *client, NMDevice *device, NMWifiP2PPeer * peer)
                        "peer", peer,
                        NULL);
 }
+
+/**
+ * nd_wfd_p2p_sink_from_uri
+ * @uri: a URI string
+ *
+ * Construct a #NdWFDP2PSink using the information encoded in the URI string
+ *
+ * Returns: The newly constructed #NdWFDP2PSink
+ */
+NdWFDP2PSink *
+nd_wfd_p2p_sink_from_uri (gchar *uri)
+{
+  GHashTable *params = nd_uri_helpers_parse_uri (uri);
+  GError *error = NULL;
+
+  /* protocol */
+  const gchar *protocol_in_uri_str = g_hash_table_lookup (params, "protocol");
+
+  NdSinkProtocol protocol_in_uri = g_ascii_strtoll (protocol_in_uri_str, NULL, 10);
+  if (protocol != protocol_in_uri)
+    {
+      g_warning ("NdWFDP2PSink: Attempted to create sink whose protocol (%s) doesn't match the URI (%s)",
+                 g_enum_to_string (ND_TYPE_SINK_PROTOCOL, protocol),
+                 g_enum_to_string (ND_TYPE_SINK_PROTOCOL, protocol_in_uri));
+      return NULL;
+    }
+
+  /* client */
+  NMClient *client = nm_client_new (NULL, &error);
+  if (!client)
+    {
+      g_warning ("NdWFDP2PSink: Failed to instantiate NetworkManager client: %s", error->message);
+      return NULL;
+    }
+
+  /* device */
+  const gchar *device_path = g_hash_table_lookup (params, "device");
+  if (!device_path)
+    {
+      g_warning ("NdWFDP2PSink: Failed to find device path in the URI %s", uri);
+      return NULL;
+    }
+  NMDevice *device = nm_client_get_device_by_path (client, device_path);
+  if (!device)
+    {
+      g_warning ("NdWFDP2PSink: Failed to find device with path %s", device_path);
+      return NULL;
+    }
+
+  /* peer */
+  const gchar *peer_path = g_hash_table_lookup (params, "peer");
+  if (!peer_path)
+    {
+      g_warning ("NdWFDP2PSink: Failed to find peer path in the URI %s", uri);
+      return NULL;
+    }
+  NMWifiP2PPeer *peer = nm_device_wifi_p2p_get_peer_by_path ((NMDeviceWifiP2P *) device, peer_path);
+  if (!peer)
+    {
+      g_warning ("NdWFDP2PSink: Failed to find peer with path %s", peer_path);
+      return NULL;
+    }
+
+  return nd_wfd_p2p_sink_new (client, device, peer);
+}
